# Hückel MO Calculation
#     Programmed by H. Kihara
import numpy as np
import PySimpleGUI as sg


sg.theme("LightGrey1")

column_buttons = [[sg.Button("Run", key="-run-", size=(12, 1)),
                sg.Button("Save to file", key="-save-", size=(14, 1)),
                sg.Button("Quit", key="-cancel-", size=(10, 1))]]

layout = [[sg.Text("Title"), sg.Input("", key="-title-")],
          [sg.Text("Number of Carbons"), sg.InputText(default_text='2', size =(4,1) , key='-natom-'),
           sg.Slider(range =(2.0, 20.0), default_value=2.0, resolution=1.0,
           orientation='h', size =(36, None) , key='-slider-', enable_events = True)],
          [sg.Text("Linear or Cyclic", size =(18,1)),
           sg.Radio("Linear Polyene", group_id="g1", key="-polyene-", default=True),
           sg.Radio("Annulene", key="-cyclopolyene-", group_id="g1")],
          [sg.Text("Charge", size =(18,1)), sg.Radio("Neutral", group_id="g2", key="-neutral-", default=True),
           sg.Radio("Cation", group_id="g2", key="-cation-"), sg.Radio("Anion", group_id="g2", key="-anion-")],
          [sg.Column(column_buttons, justification='r')],
          [sg.Listbox([], size=(700, 100), font=('Courier New', 11), enable_events=False, key="-LIST-")]]

window = sg.Window("Hückel MO Calculation", layout, font=(None,14), size=(740,640))

# Check the data
def data_check(values) :
    n_atom = int(values["-natom-"] )

    if n_atom < 3 and values["-cyclopolyene-"] == True:
        sg.PopupTimed("Cyclic cannot be specified when the number of carbons is less than 3", no_titlebar=True)
        return True

    return False

# Get data and constructs Secular Determinant
def secdet(values) :
    title = values["-title-"]

    n_atom =  int(values["-natom-"] )

    # Linear or Cyclic
    if values["-polyene-"] == True:
        n_bond = n_atom - 1
    else:
        n_bond = n_atom

    charge = 0
    if values["-cation-"] == True:
        charge = 1
    if values["-anion-"] == True:
        charge = -1

    n_elec = n_atom - charge

    info = f" {n_atom:>2} atoms   {n_bond:>2} bonds  {n_elec:>2} electrons"

    H = np.zeros((n_atom, n_atom))

    for i in range(n_atom-1):
        H[i, i+1] = 1.0
        H[i+1, i] = 1.0

    if values["-cyclopolyene-"] == True:
        H[0, n_atom-1] = 1.0
        H[n_atom-1, 0] = 1.0

    return title, info, n_atom, n_elec, H

def diag(H):
    """
        Diagonalize matrix H and using numpy.linalg.eigh
        Sort eigenvalues and eigenvectors in descending order of eigenvalues

        Parameter
            H : Two-dimensional matrix of floating point number
                Hamiltonian
        Return
            E : Vector of floating point number
                Eigen values
            v : Two-dimensional matrix of floating point number
                Eigen vectors, each vector as a row vector
                [[c11, c12, ..., c1n], [c21, c22, ..., c2n], ...  [cn1, cn2, ..., cnn]]
    """
    p = H.shape[0]
    E, v = np.linalg.eigh(H)

    sort_E = np.empty(p)
    for i in range(p):
        sort_E[i] = E[i]

    idx = np.argsort(sort_E)[::-1]    # Reverse index and sort in descending order
    E = E[idx]
    v = v.T[idx]

    return E, v

def output(title, info, n_atom, E, v, p):
    column = 7
    result = [title]
    result.append("")
    result.append(info)
    result.append("")
    result.append("--  Eigenvalues and Eigenvector   --")

    eig_print(result, column, n_atom, E, v)

    pmat_print(result, column, n_atom, p)

    window['-LIST-'].update(result)

    return result

def eig_print(result, column, n_atom, E, v):
    first = 0
    last = first + column

    while first < n_atom :
        result.append("")
        if last > n_atom :
            last = n_atom
        txt = ''
        for i in range(first, last):
            txt += f"{i+1:>10}"

        result.append(txt)

        txt = '  '
        for i in range(first, last):
            txt += f"{E[i]:>10.4f}"

        result.append(txt)
        result.append("")

        for i in range(n_atom):
            txt = f"{i+1:>2}"
            for j in range(first, last):
                txt += f"{v[j][i]:>10.4f}"

            result.append(txt)

        result.append("")
        first = last
        last = first + column

# Calculates bond order
def calc_pmat(E, v, n, n_elec):
    odd = False
    degenerate = False
    EPS = 0.000001

    nocc = (n_elec - 1) // 2      # Number of occupied orbitals
    if n_elec % 2 != 0:
        odd = True

    if (E[nocc] - E[nocc+1]) < EPS:    # Check if HOMO is degenerate
        degenerate = True

    p = np.zeros((n, n))

    for i in range(n):
        for j in range(i+1):
            for k in range(nocc):
                p[i][j]= p[i][j] + 2.0 * v[k][i] * v[k][j]

            if degenerate:
                if odd:
                    p[i][j] = p[i][j] + 0.5 * v[nocc][i]* v[nocc][j] + v[nocc+1][i] * v[nocc+1][j]
                else:
                    p[i][j] = p[i][j] + v[nocc][i]* v[nocc][j] + v[nocc+1][i] * v[nocc+1][j]

            else:
                if odd:
                    p[i][j]= p[i][j] + v[nocc][i] * v[nocc][j]
                else:
                    p[i][j]= p[i][j]+ 2.0 * v[nocc][i] * v[nocc][j]

    return p

# Print bond order matrix
def pmat_print(result, column, n, p):

    result.append("-- Bond Order Matrix --")
    result.append("")
    first = 0
    last = first + column

    while first < n :
        txt = ""

        if last > n :
            last = n
        for i in range(first, last):
            txt += f"{i+1:>10}"

        result.append(txt)

        for i in range(n):
            txt = f"{i+1:>2}"
            for j in range(first, last):
                txt += f'{p[i][j]:>10.4f}'

            result.append(txt)

        result.append("")
        first = last
        last = first + column

def save_file(result):
    if len(result) > 0:
        savename = sg.popup_get_file("Specify a file to save the results", save_as=True, no_titlebar=True)
        if not savename:
            sg.PopupTimed("Please enter a file name")
            return

        fw = open(savename, 'w')

        for i in range(len(result)):
            fw.write(result[i]+"\n")

        fw.close

#  =========  Main  ==========

while True:
    event, values = window.read()

    if event == "-run-":
        ErrorOccurred = data_check(values)

        if not ErrorOccurred:
            title, info, n_atom, n_elec, H = secdet(values)        # Construction of secular determinant

            E, v = diag(H)      # Diagonalization

            p = calc_pmat(E, v, n_atom, n_elec)        # Calculation of bond order

            result = output(title, info, n_atom, E, v, p)        # Show results

    if event == '-slider-':
        window ['-natom-'].Update(int(values['-slider-']))

    if event == "-save-":
        save_file(result)

    if event == sg.WIN_CLOSED or event == "-cancel-":
        break

    if event == None:
        break

window.close()
